# Generated File
from __future__ import annotations

import dataclasses
import struct
import typing
import typing_extensions

from retro_data_structures import json_util
from retro_data_structures.game_check import Game
from retro_data_structures.properties.base_property import BaseObjectType
from retro_data_structures.properties.field_reflection import FieldReflection
import retro_data_structures.enums.corruption as enums
from retro_data_structures.properties.corruption.archetypes.BonusCredit import BonusCredit
from retro_data_structures.properties.corruption.archetypes.EditorProperties import EditorProperties
from retro_data_structures.properties.corruption.core.AssetId import AssetId, default_asset_id

if typing.TYPE_CHECKING:
    class AchievementJson(typing_extensions.TypedDict):
        editor_properties: json_util.JsonObject
        bonus_credit: json_util.JsonObject
        achievement: int
        normal_difficulty: bool
        hard_difficulty: bool
        elite_difficulty: bool
        bonus_credit_string: int
    

@dataclasses.dataclass()
class Achievement(BaseObjectType):
    editor_properties: EditorProperties = dataclasses.field(default_factory=EditorProperties, metadata={
        'reflection': FieldReflection[EditorProperties](
            EditorProperties, id=0x255a4580, original_name='EditorProperties', from_json=EditorProperties.from_json, to_json=EditorProperties.to_json
        ),
    })
    bonus_credit: BonusCredit = dataclasses.field(default_factory=BonusCredit, metadata={
        'reflection': FieldReflection[BonusCredit](
            BonusCredit, id=0x7aac9e22, original_name='BonusCredit', from_json=BonusCredit.from_json, to_json=BonusCredit.to_json
        ),
    })
    achievement: enums.Achievement = dataclasses.field(default=enums.Achievement.Unknown84, metadata={
        'reflection': FieldReflection[enums.Achievement](
            enums.Achievement, id=0x058d2ddb, original_name='Achievement', from_json=enums.Achievement.from_json, to_json=enums.Achievement.to_json
        ),
    })
    normal_difficulty: bool = dataclasses.field(default=True, metadata={
        'reflection': FieldReflection[bool](
            bool, id=0x974f4aa1, original_name='NormalDifficulty'
        ),
    })
    hard_difficulty: bool = dataclasses.field(default=True, metadata={
        'reflection': FieldReflection[bool](
            bool, id=0x0f8cf6ff, original_name='HardDifficulty'
        ),
    })
    elite_difficulty: bool = dataclasses.field(default=True, metadata={
        'reflection': FieldReflection[bool](
            bool, id=0x9b8903eb, original_name='EliteDifficulty'
        ),
    })
    bonus_credit_string: AssetId = dataclasses.field(default=default_asset_id, metadata={
        'asset_types': ['STRG'], 'reflection': FieldReflection[AssetId](
            AssetId, id=0xd6a0cff1, original_name='BonusCreditString'
        ),
    })

    @classmethod
    def game(cls) -> Game:
        return Game.CORRUPTION

    def get_name(self) -> str | None:
        return self.editor_properties.name

    def set_name(self, name: str) -> None:
        self.editor_properties.name = name

    @classmethod
    def object_type(cls) -> str:
        return 'ACHI'

    @classmethod
    def modules(cls) -> typing.List[str]:
        return ['RSO_ScriptAchievement.rso']

    @classmethod
    def from_stream(cls, data: typing.BinaryIO, size: int | None = None, default_override: dict | None = None) -> typing_extensions.Self:
        struct_id, size, property_count = struct.unpack(">LHH", data.read(8))
        assert struct_id == 0xFFFFFFFF
        root_size_start = data.tell() - 2

        if (result := cls._fast_decode(data, property_count)) is not None:
            return result

        present_fields = default_override or {}
        for _ in range(property_count):
            property_id, property_size = struct.unpack(">LH", data.read(6))
            start = data.tell()
            try:
                property_name, decoder = _property_decoder[property_id]
                present_fields[property_name] = decoder(data, property_size)
            except KeyError:
                raise RuntimeError(f"Unknown property: 0x{property_id:08x}")
            assert data.tell() - start == property_size

        assert data.tell() - root_size_start == size
        return cls(**present_fields)

    @classmethod
    def _fast_decode(cls, data: typing.BinaryIO, property_count: int) -> typing_extensions.Self | None:
        if property_count != 7:
            return None
    
        property_id, property_size = struct.unpack(">LH", data.read(6))
        assert property_id == 0x255a4580
        editor_properties = EditorProperties.from_stream(data, property_size)
    
        property_id, property_size = struct.unpack(">LH", data.read(6))
        assert property_id == 0x7aac9e22
        bonus_credit = BonusCredit.from_stream(data, property_size)
    
        property_id, property_size = struct.unpack(">LH", data.read(6))
        assert property_id == 0x058d2ddb
        achievement = enums.Achievement.from_stream(data)
    
        property_id, property_size = struct.unpack(">LH", data.read(6))
        assert property_id == 0x974f4aa1
        normal_difficulty = struct.unpack('>?', data.read(1))[0]
    
        property_id, property_size = struct.unpack(">LH", data.read(6))
        assert property_id == 0x0f8cf6ff
        hard_difficulty = struct.unpack('>?', data.read(1))[0]
    
        property_id, property_size = struct.unpack(">LH", data.read(6))
        assert property_id == 0x9b8903eb
        elite_difficulty = struct.unpack('>?', data.read(1))[0]
    
        property_id, property_size = struct.unpack(">LH", data.read(6))
        assert property_id == 0xd6a0cff1
        bonus_credit_string = struct.unpack(">Q", data.read(8))[0]
    
        return cls(editor_properties, bonus_credit, achievement, normal_difficulty, hard_difficulty, elite_difficulty, bonus_credit_string)

    def to_stream(self, data: typing.BinaryIO, default_override: dict | None = None) -> None:
        default_override = default_override or {}
        data.write(b'\xff\xff\xff\xff')  # struct object id
        root_size_offset = data.tell()
        data.write(b'\x00\x00')  # placeholder for root struct size
        data.write(b'\x00\x07')  # 7 properties

        data.write(b'%ZE\x80')  # 0x255a4580
        before = data.tell()
        data.write(b'\x00\x00')  # size placeholder
        self.editor_properties.to_stream(data)
        after = data.tell()
        data.seek(before)
        data.write(struct.pack(">H", after - before - 2))
        data.seek(after)

        data.write(b'z\xac\x9e"')  # 0x7aac9e22
        before = data.tell()
        data.write(b'\x00\x00')  # size placeholder
        self.bonus_credit.to_stream(data)
        after = data.tell()
        data.seek(before)
        data.write(struct.pack(">H", after - before - 2))
        data.seek(after)

        data.write(b'\x05\x8d-\xdb')  # 0x58d2ddb
        data.write(b'\x00\x04')  # size
        self.achievement.to_stream(data)

        data.write(b'\x97OJ\xa1')  # 0x974f4aa1
        data.write(b'\x00\x01')  # size
        data.write(struct.pack('>?', self.normal_difficulty))

        data.write(b'\x0f\x8c\xf6\xff')  # 0xf8cf6ff
        data.write(b'\x00\x01')  # size
        data.write(struct.pack('>?', self.hard_difficulty))

        data.write(b'\x9b\x89\x03\xeb')  # 0x9b8903eb
        data.write(b'\x00\x01')  # size
        data.write(struct.pack('>?', self.elite_difficulty))

        data.write(b'\xd6\xa0\xcf\xf1')  # 0xd6a0cff1
        data.write(b'\x00\x08')  # size
        data.write(struct.pack(">Q", self.bonus_credit_string))

        struct_end_offset = data.tell()
        data.seek(root_size_offset)
        data.write(struct.pack(">H", struct_end_offset - root_size_offset - 2))
        data.seek(struct_end_offset)

    @classmethod
    def from_json(cls, data: json_util.JsonValue) -> typing_extensions.Self:
        json_data = typing.cast("AchievementJson", data)
        return cls(
            editor_properties=EditorProperties.from_json(json_data['editor_properties']),
            bonus_credit=BonusCredit.from_json(json_data['bonus_credit']),
            achievement=enums.Achievement.from_json(json_data['achievement']),
            normal_difficulty=json_data['normal_difficulty'],
            hard_difficulty=json_data['hard_difficulty'],
            elite_difficulty=json_data['elite_difficulty'],
            bonus_credit_string=json_data['bonus_credit_string'],
        )

    def to_json(self) -> json_util.JsonObject:
        return {
            'editor_properties': self.editor_properties.to_json(),
            'bonus_credit': self.bonus_credit.to_json(),
            'achievement': self.achievement.to_json(),
            'normal_difficulty': self.normal_difficulty,
            'hard_difficulty': self.hard_difficulty,
            'elite_difficulty': self.elite_difficulty,
            'bonus_credit_string': self.bonus_credit_string,
        }


def _decode_achievement(data: typing.BinaryIO, property_size: int):
    return enums.Achievement.from_stream(data)


def _decode_normal_difficulty(data: typing.BinaryIO, property_size: int):
    return struct.unpack('>?', data.read(1))[0]


def _decode_hard_difficulty(data: typing.BinaryIO, property_size: int):
    return struct.unpack('>?', data.read(1))[0]


def _decode_elite_difficulty(data: typing.BinaryIO, property_size: int):
    return struct.unpack('>?', data.read(1))[0]


def _decode_bonus_credit_string(data: typing.BinaryIO, property_size: int):
    return struct.unpack(">Q", data.read(8))[0]


_property_decoder: typing.Dict[int, typing.Tuple[str, typing.Callable[[typing.BinaryIO, int], typing.Any]]] = {
    0x255a4580: ('editor_properties', EditorProperties.from_stream),
    0x7aac9e22: ('bonus_credit', BonusCredit.from_stream),
    0x58d2ddb: ('achievement', _decode_achievement),
    0x974f4aa1: ('normal_difficulty', _decode_normal_difficulty),
    0xf8cf6ff: ('hard_difficulty', _decode_hard_difficulty),
    0x9b8903eb: ('elite_difficulty', _decode_elite_difficulty),
    0xd6a0cff1: ('bonus_credit_string', _decode_bonus_credit_string),
}
