# Generated File
from __future__ import annotations

import dataclasses
import struct
import typing
import typing_extensions

from retro_data_structures import json_util
from retro_data_structures.game_check import Game
from retro_data_structures.properties.base_property import BaseObjectType
from retro_data_structures.properties.field_reflection import FieldReflection
from retro_data_structures.properties.corruption.archetypes.EditorProperties import EditorProperties
from retro_data_structures.properties.corruption.archetypes.ScannableParameters import ScannableParameters
from retro_data_structures.properties.corruption.core.Vector import Vector

if typing.TYPE_CHECKING:
    class PointOfInterestJson(typing_extensions.TypedDict):
        editor_properties: json_util.JsonObject
        scan_info: json_util.JsonObject
        scan_offset: json_util.JsonValue
        look_at_poi: bool
        scan_offset_local: bool
    

@dataclasses.dataclass()
class PointOfInterest(BaseObjectType):
    editor_properties: EditorProperties = dataclasses.field(default_factory=EditorProperties, metadata={
        'reflection': FieldReflection[EditorProperties](
            EditorProperties, id=0x255a4580, original_name='EditorProperties', from_json=EditorProperties.from_json, to_json=EditorProperties.to_json
        ),
    })
    scan_info: ScannableParameters = dataclasses.field(default_factory=ScannableParameters, metadata={
        'reflection': FieldReflection[ScannableParameters](
            ScannableParameters, id=0xbdbec295, original_name='ScanInfo', from_json=ScannableParameters.from_json, to_json=ScannableParameters.to_json
        ),
    })
    scan_offset: Vector = dataclasses.field(default_factory=lambda: Vector(x=0.0, y=0.0, z=0.0), metadata={
        'reflection': FieldReflection[Vector](
            Vector, id=0x863c16e3, original_name='ScanOffset', from_json=Vector.from_json, to_json=Vector.to_json
        ),
    })
    look_at_poi: bool = dataclasses.field(default=False, metadata={
        'reflection': FieldReflection[bool](
            bool, id=0x01f9c5bb, original_name='Look At POI'
        ),
    })
    scan_offset_local: bool = dataclasses.field(default=True, metadata={
        'reflection': FieldReflection[bool](
            bool, id=0x91351b69, original_name='ScanOffsetLocal'
        ),
    })

    @classmethod
    def game(cls) -> Game:
        return Game.CORRUPTION

    def get_name(self) -> str | None:
        return self.editor_properties.name

    def set_name(self, name: str) -> None:
        self.editor_properties.name = name

    @classmethod
    def object_type(cls) -> str:
        return 'POIN'

    @classmethod
    def from_stream(cls, data: typing.BinaryIO, size: int | None = None, default_override: dict | None = None) -> typing_extensions.Self:
        struct_id, size, property_count = struct.unpack(">LHH", data.read(8))
        assert struct_id == 0xFFFFFFFF
        root_size_start = data.tell() - 2

        if (result := cls._fast_decode(data, property_count)) is not None:
            return result

        present_fields = default_override or {}
        for _ in range(property_count):
            property_id, property_size = struct.unpack(">LH", data.read(6))
            start = data.tell()
            try:
                property_name, decoder = _property_decoder[property_id]
                present_fields[property_name] = decoder(data, property_size)
            except KeyError:
                raise RuntimeError(f"Unknown property: 0x{property_id:08x}")
            assert data.tell() - start == property_size

        assert data.tell() - root_size_start == size
        return cls(**present_fields)

    @classmethod
    def _fast_decode(cls, data: typing.BinaryIO, property_count: int) -> typing_extensions.Self | None:
        if property_count != 5:
            return None
    
        property_id, property_size = struct.unpack(">LH", data.read(6))
        assert property_id == 0x255a4580
        editor_properties = EditorProperties.from_stream(data, property_size)
    
        property_id, property_size = struct.unpack(">LH", data.read(6))
        assert property_id == 0xbdbec295
        scan_info = ScannableParameters.from_stream(data, property_size)
    
        property_id, property_size = struct.unpack(">LH", data.read(6))
        assert property_id == 0x863c16e3
        scan_offset = Vector.from_stream(data)
    
        property_id, property_size = struct.unpack(">LH", data.read(6))
        assert property_id == 0x01f9c5bb
        look_at_poi = struct.unpack('>?', data.read(1))[0]
    
        property_id, property_size = struct.unpack(">LH", data.read(6))
        assert property_id == 0x91351b69
        scan_offset_local = struct.unpack('>?', data.read(1))[0]
    
        return cls(editor_properties, scan_info, scan_offset, look_at_poi, scan_offset_local)

    def to_stream(self, data: typing.BinaryIO, default_override: dict | None = None) -> None:
        default_override = default_override or {}
        data.write(b'\xff\xff\xff\xff')  # struct object id
        root_size_offset = data.tell()
        data.write(b'\x00\x00')  # placeholder for root struct size
        data.write(b'\x00\x05')  # 5 properties

        data.write(b'%ZE\x80')  # 0x255a4580
        before = data.tell()
        data.write(b'\x00\x00')  # size placeholder
        self.editor_properties.to_stream(data)
        after = data.tell()
        data.seek(before)
        data.write(struct.pack(">H", after - before - 2))
        data.seek(after)

        data.write(b'\xbd\xbe\xc2\x95')  # 0xbdbec295
        before = data.tell()
        data.write(b'\x00\x00')  # size placeholder
        self.scan_info.to_stream(data)
        after = data.tell()
        data.seek(before)
        data.write(struct.pack(">H", after - before - 2))
        data.seek(after)

        data.write(b'\x86<\x16\xe3')  # 0x863c16e3
        data.write(b'\x00\x0c')  # size
        self.scan_offset.to_stream(data)

        data.write(b'\x01\xf9\xc5\xbb')  # 0x1f9c5bb
        data.write(b'\x00\x01')  # size
        data.write(struct.pack('>?', self.look_at_poi))

        data.write(b'\x915\x1bi')  # 0x91351b69
        data.write(b'\x00\x01')  # size
        data.write(struct.pack('>?', self.scan_offset_local))

        struct_end_offset = data.tell()
        data.seek(root_size_offset)
        data.write(struct.pack(">H", struct_end_offset - root_size_offset - 2))
        data.seek(struct_end_offset)

    @classmethod
    def from_json(cls, data: json_util.JsonValue) -> typing_extensions.Self:
        json_data = typing.cast("PointOfInterestJson", data)
        return cls(
            editor_properties=EditorProperties.from_json(json_data['editor_properties']),
            scan_info=ScannableParameters.from_json(json_data['scan_info']),
            scan_offset=Vector.from_json(json_data['scan_offset']),
            look_at_poi=json_data['look_at_poi'],
            scan_offset_local=json_data['scan_offset_local'],
        )

    def to_json(self) -> json_util.JsonObject:
        return {
            'editor_properties': self.editor_properties.to_json(),
            'scan_info': self.scan_info.to_json(),
            'scan_offset': self.scan_offset.to_json(),
            'look_at_poi': self.look_at_poi,
            'scan_offset_local': self.scan_offset_local,
        }


def _decode_scan_offset(data: typing.BinaryIO, property_size: int):
    return Vector.from_stream(data)


def _decode_look_at_poi(data: typing.BinaryIO, property_size: int):
    return struct.unpack('>?', data.read(1))[0]


def _decode_scan_offset_local(data: typing.BinaryIO, property_size: int):
    return struct.unpack('>?', data.read(1))[0]


_property_decoder: typing.Dict[int, typing.Tuple[str, typing.Callable[[typing.BinaryIO, int], typing.Any]]] = {
    0x255a4580: ('editor_properties', EditorProperties.from_stream),
    0xbdbec295: ('scan_info', ScannableParameters.from_stream),
    0x863c16e3: ('scan_offset', _decode_scan_offset),
    0x1f9c5bb: ('look_at_poi', _decode_look_at_poi),
    0x91351b69: ('scan_offset_local', _decode_scan_offset_local),
}
